package net.geant.nmaas.nmservice.deployment.containerorchestrators.kubernetes.components.janitor;

import io.grpc.ConnectivityState;
import io.grpc.ManagedChannel;
import io.grpc.ManagedChannelBuilder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import net.geant.nmaas.externalservices.inventory.janitor.BasicAuthServiceGrpc;
import net.geant.nmaas.externalservices.inventory.janitor.CertManagerServiceGrpc;
import net.geant.nmaas.externalservices.inventory.janitor.ConfigServiceGrpc;
import net.geant.nmaas.externalservices.inventory.janitor.InformationServiceGrpc;
import net.geant.nmaas.externalservices.inventory.janitor.JanitorManager;
import net.geant.nmaas.externalservices.inventory.janitor.NamespaceServiceGrpc;
import net.geant.nmaas.externalservices.inventory.janitor.PodServiceGrpc;
import net.geant.nmaas.externalservices.inventory.janitor.ReadinessServiceGrpc;
import net.geant.nmaas.externalservices.kubernetes.KubernetesClusterNamespaceService;
import net.geant.nmaas.orchestration.Identifier;
import net.geant.nmaas.portal.api.domain.KeyValueView;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

@Service
@RequiredArgsConstructor
@Slf4j
public class JanitorService {

    private final KubernetesClusterNamespaceService namespaceService;
    private final ManagedChannel channel;

    @Autowired
    public JanitorService(KubernetesClusterNamespaceService namespaceService, Environment env) {
        this.namespaceService = namespaceService;
        this.channel = ManagedChannelBuilder.forAddress(
                env.getProperty("janitor.address"),
                env.getProperty("janitor.port", Integer.class))
                .maxInboundMessageSize(Integer.MAX_VALUE)
                .usePlaintext()
                .build();
    }

    private JanitorManager.InstanceRequest buildInstanceRequest(Identifier deploymentId, String domain) {
        JanitorManager.Instance instance = JanitorManager.Instance.newBuilder().
                setNamespace(namespaceService.namespace(domain)).
                setUid(deploymentId.value()).
                setDomain(domain).
                build();
        return JanitorManager.InstanceRequest.newBuilder().
                setApi("v1").
                setDeployment(instance).
                build();
    }

    private JanitorManager.InstanceCredentialsRequest buildInstanceCredentialsRequest(Identifier deploymentId, String domain, String user, String password) {
        JanitorManager.Instance instance = JanitorManager.Instance.newBuilder().
                setNamespace(namespaceService.namespace(domain)).
                setUid(deploymentId.value()).
                setDomain(domain).
                build();
        JanitorManager.Credentials credentials = JanitorManager.Credentials.newBuilder().
                setUser(user).
                setPassword(password).
                build();
        return JanitorManager.InstanceCredentialsRequest.newBuilder().
                setApi("v1").
                setInstance(instance).
                setCredentials(credentials).
                build();
    }

    private JanitorManager.NamespaceRequest buildNamespaceRequest(String domain, List<KeyValueView> annotations) {
        return JanitorManager.NamespaceRequest.newBuilder()
                .setApi("v1")
                .setNamespace(domain)
                .addAllAnnotations(annotations.stream()
                        .map(kv -> JanitorManager.KeyValue.newBuilder().setKey(kv.getKey()).setValue(kv.getValue()).build())
                        .collect(Collectors.toList()))
                .build();
    }

    public void createOrReplaceConfigMap(Identifier deploymentId, String domain) {
        log.info("Creating or replacing configMap(s) for deployment {} in domain {}", deploymentId.value(), domain);
        ConfigServiceGrpc.ConfigServiceBlockingStub stub = ConfigServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.createOrReplace(buildInstanceRequest(deploymentId, domain));
        throwExceptionIfExecutionFailed(response);
    }

    public void deleteConfigMapIfExists(Identifier deploymentId, String domain) {
        log.info("Deleting configMap(s) for deployment {} in domain {}", deploymentId.value(), domain);
        ConfigServiceGrpc.ConfigServiceBlockingStub stub = ConfigServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.deleteIfExists(buildInstanceRequest(deploymentId, domain));
        throwExceptionIfExecutionFailed(response);
    }

    public void createOrReplaceBasicAuth(Identifier deploymentId, String domain, String user, String password) {
        log.info("Configuring basic auth for deployment {} in domain {}", deploymentId.value(), domain);
        BasicAuthServiceGrpc.BasicAuthServiceBlockingStub stub = BasicAuthServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.createOrReplace(buildInstanceCredentialsRequest(deploymentId, domain, user, password));
        throwExceptionIfExecutionFailed(response);
    }

    public void deleteBasicAuthIfExists(Identifier deploymentId, String domain) {
        log.info("Deleting basic auth for deployment {} in domain {}", deploymentId.value(), domain);
        BasicAuthServiceGrpc.BasicAuthServiceBlockingStub stub = BasicAuthServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.deleteIfExists(buildInstanceRequest(deploymentId, domain));
        throwExceptionIfExecutionFailed(response);
    }

    public void deleteTlsIfExists(Identifier deploymentId, String domain) {
        log.info("Deleting TLS for deployment {} in domain {}", deploymentId.value(), domain);
        CertManagerServiceGrpc.CertManagerServiceBlockingStub stub = CertManagerServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.deleteIfExists(buildInstanceRequest(deploymentId, domain));
        throwExceptionIfExecutionFailed(response);
    }

    private void throwExceptionIfExecutionFailed(JanitorManager.ServiceResponse response) {
        if (response.getStatus() != JanitorManager.Status.OK) {
            throw new JanitorResponseException(janitorExceptionMessage(response.getMessage()));
        }
    }

    boolean isJanitorAvailable(){
        return Arrays.asList(ConnectivityState.CONNECTING, ConnectivityState.IDLE, ConnectivityState.READY).contains(this.channel.getState(false));
    }

    public boolean checkIfReady(Identifier deploymentId, String domain) {
        log.trace("Checking if deployment {} in domain {} is ready", deploymentId.value(), domain);
        ReadinessServiceGrpc.ReadinessServiceBlockingStub stub = ReadinessServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.checkIfReady(buildInstanceRequest(deploymentId, domain));
        switch (response.getStatus()) {
            case OK:
                return true;
            case PENDING:
                return false;
            case FAILED:
            case UNRECOGNIZED:
            default:
                throw new JanitorResponseException(janitorExceptionMessage(response.getMessage()));
        }
    }

    public String retrieveServiceIp(Identifier serviceId, String domain) {
        log.info("Retrieving service IP for {} in domain {}", serviceId.value(), domain);
        InformationServiceGrpc.InformationServiceBlockingStub stub = InformationServiceGrpc.newBlockingStub(channel);
        JanitorManager.InfoServiceResponse response = stub.retrieveServiceIp(buildInstanceRequest(serviceId, domain));
        switch (response.getStatus()) {
            case OK:
                return response.getInfo();
            case FAILED:
            default:
                throw new JanitorResponseException(janitorExceptionMessage(response.getMessage()));
        }
    }

    public void checkServiceExists(Identifier serviceId, String domain) {
        log.info("Verifying if provided service {} exists in domain {}", serviceId.value(), domain);
        InformationServiceGrpc.InformationServiceBlockingStub stub = InformationServiceGrpc.newBlockingStub(channel);
        JanitorManager.InfoServiceResponse response = stub.checkServiceExists(buildInstanceRequest(serviceId, domain));
        switch (response.getStatus()) {
            case OK:
                return;
            case FAILED:
            default:
                throw new JanitorResponseException(janitorExceptionMessage(response.getMessage()));
        }
    }

    public List<JanitorManager.PodInfo> getPodNames(Identifier deploymentId, String domain) {
        log.debug("Retrieving list of pods for {} in domain {}", deploymentId.value(), domain);
        PodServiceGrpc.PodServiceBlockingStub stub = PodServiceGrpc.newBlockingStub(channel);
        JanitorManager.PodListResponse response = stub.retrievePodList(buildInstanceRequest(deploymentId, domain));
        switch (response.getStatus()) {
            case OK:
                return response.getPodsList();
            case FAILED:
            default:
                throw new JanitorResponseException(janitorExceptionMessage(response.getMessage()));
        }
    }

    public List<String> getPodLogs(Identifier deploymentId, String podName, String containerName, String domain) {
        PodServiceGrpc.PodServiceBlockingStub stub = PodServiceGrpc.newBlockingStub(channel);
        JanitorManager.PodInfo podInfo = (StringUtils.isNotEmpty(containerName)) ?
                JanitorManager.PodInfo.newBuilder().setName(podName).setDisplayName(podName).addContainers(containerName).build() :
                JanitorManager.PodInfo.newBuilder().setName(podName).setDisplayName(podName).build();
        JanitorManager.PodLogsResponse response = stub.retrievePodLogs(
                JanitorManager.PodRequest.newBuilder()
                        .setApi("v1")
                        .setDeployment(
                                JanitorManager.Instance.newBuilder().
                                        setNamespace(namespaceService.namespace(domain)).
                                        setUid(deploymentId.value()).
                                        setDomain(domain).build()
                        )
                        .setPod(podInfo)
                        .build());
        switch (response.getStatus()) {
            case OK:
                return response.getLinesList();
            case FAILED:
            default:
                throw new JanitorResponseException(janitorExceptionMessage(response.getMessage()));
        }
    }

    public void createNameSpace(String domainNameSpace, List<KeyValueView> annotations) {
        log.info("Request domain namespace creation for domain {} with {} annotations", domainNameSpace, annotations.size());
        NamespaceServiceGrpc.NamespaceServiceBlockingStub stub = NamespaceServiceGrpc.newBlockingStub(channel);
        JanitorManager.ServiceResponse response = stub.createNamespace(
                buildNamespaceRequest(domainNameSpace, annotations));
        throwExceptionIfExecutionFailed(response);
    }

    private static String janitorExceptionMessage(String message) {
        return "Error response from Janitor: " + message;
    }

}
