package net.geant.nmaas.portal.api.user;

import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import lombok.extern.slf4j.Slf4j;
import net.geant.nmaas.portal.api.domain.SSHKeyRequest;
import net.geant.nmaas.portal.api.domain.SSHKeyView;
import net.geant.nmaas.portal.persistent.entity.User;
import net.geant.nmaas.portal.service.SSHKeyService;
import net.geant.nmaas.portal.service.UserService;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import jakarta.validation.Valid;
import java.security.Principal;
import java.util.List;

import static net.geant.nmaas.portal.persistent.entity.Role.ROLE_SYSTEM_ADMIN;

@RestController
@RequestMapping("/api")
@RequiredArgsConstructor
@Slf4j
public class SSHKeysController {

    private final SSHKeyService keysService;
    private final UserService userService;

    @GetMapping("/user/keys")
    public List<SSHKeyView> getAllByUser(Principal principal) {
        User owner = this.getUser(principal);
        return this.keysService.findAllByUser(owner);
    }

    @PutMapping("/user/keys")
    public void create(Principal principal, @RequestBody @Valid SSHKeyRequest request) {
        User owner = this.getUser(principal);
        this.keysService.create(request, owner);
    }

    @DeleteMapping("/user/keys/{id}")
    public void invalidate(Principal principal, @PathVariable Long id) {
        User owner = this.getUser(principal);
        this.keysService.invalidate(owner, id);
    }

    @PreAuthorize("hasRole('ROLE_SYSTEM_ADMIN')")
    @GetMapping("/user/keys/view/{id}")
    public List<SSHKeyView> getAllByUserId(Principal principal, @PathVariable Long id) {
        User requester = this.getUser(principal);
        if(!isAdmin(requester)){
            throw new IllegalArgumentException("You need admin privileges to view user keys.");
        }
        User user = this.getUser(id);
        return this.keysService.findAllByUser(user);
    }

    @PreAuthorize("hasRole('ROLE_SYSTEM_ADMIN')")
    @PutMapping("/user/keys/view/{id}")
    public void createUserKey(Principal principal, @PathVariable Long id, @RequestBody @Valid SSHKeyRequest request) {
        User requester = this.getUser(principal);
        if(!isAdmin(requester)){
            throw new IllegalArgumentException("You need admin privileges to edit user keys.");
        }
        User user = this.getUser(id);
        this.keysService.create(request, user);
    }

    @PreAuthorize("hasRole('ROLE_SYSTEM_ADMIN')")
    @DeleteMapping("/user/keys/view/{userId}/{keyId}")
    public void invalidateUserKey(Principal principal, @PathVariable Long keyId, @PathVariable Long userId) {
        User requester = this.getUser(principal);
        if(!isAdmin(requester)){
            throw new IllegalArgumentException("You need admin privileges to delete user keys.");
        }
        User owner = this.getUser(userId);
        this.keysService.invalidate(owner, keyId);
    }


    private User getUser(Principal principal) {
        return this.userService.findByUsername(principal.getName()).orElseThrow(() -> new IllegalArgumentException("User not found"));
    }

    private User getUser(Long id) {
        return this.userService.findById(id).orElseThrow(() -> new IllegalArgumentException("User not found"));
    }

    private boolean isAdmin(User user) {
        return user.getRoles().stream().anyMatch(role -> role.getRole().equals(ROLE_SYSTEM_ADMIN));
    }

}
